from time import perf_counter

from .log import getLogger


def measureTime(func):
    def wrapper(*args, **kwargs):
        start_time = perf_counter()
        error = None
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            error = e
        end_time = perf_counter()
        getLogger().info(
            f"Function {func.__name__!r} took {end_time - start_time} seconds to run"
        )
        if error:
            raise error
        return result

    return wrapper



from typing import Any, Callable, Sequence, Union, Iterable, Optional
from queue import LifoQueue
import re

__all__ = [
    "evalToBasicType",
    "startswith",
    "LineIterator",
]


def evalToBasicType(s: str) -> Union[str, int, float]:
    """
    Evaluate a string to a basic type (str, int, float)
    """
    if not isinstance(s, str):
        raise TypeError("Input must be a string")
    s = s.strip()
    if re.fullmatch(r"\"[^\"]*\"", s):
        # "string"
        value = s[1:-1]
    elif re.fullmatch(r"\'[^']*\'", s):
        # 'string'
        value = s[1:-1]
    elif re.fullmatch(r"\d+", s):
        # 123
        value = int(s)
    elif re.fullmatch(r"\d*\.\d+|\d+\.", s):
        # 1.23
        # .23
        # 1.
        value = float(s)
    else:
        value = s
    return value


def startswith(
    string: str,
    patterns: Sequence[str],
    regex: bool = False,
    case: bool = False,
) -> Union[re.Match, bool]:
    if regex:
        flags = re.IGNORECASE if case else 0
        for ptn in patterns:
            obj = re.match(ptn, string, flags)
            if obj:
                return obj
    else:
        string = string.strip()
        for ptn in patterns:
            if string.startswith(str(ptn)):
                return True
    return False


class LineIterator:
    status: Any = None

    def __init__(
        self,
        data: Iterable[str],
        partition: Optional[str] = None,
        chomp: bool = False,
        cb_init: Optional[Callable] = None,
        cb_next: Optional[Callable] = None,
    ) -> None:
        if not isinstance(data, Iterable):
            raise TypeError(f"data should be an iterable - {repr(data)}")
        self.last: list = []
        self.last1: str = ""
        self.line: int = 0
        self.partition = partition
        self.chomp = chomp
        self.__reg: LifoQueue = LifoQueue()
        data_sequence = tuple(data)
        self.total_lines = len(data_sequence)
        self.__data_iter = iter(data_sequence)
        if isinstance(cb_init, Callable):
            cb_init(self)
        self.__cb_next = cb_next

    def __str__(self) -> str:
        return self.last1

    @property
    def next(self) -> str:
        if not self.__reg.empty():
            data = self.__reg.get()
        else:
            data = next(self.__data_iter)
            if self.chomp:
                data = re.sub(r"\n$", "", data)
        self.last.append(data)
        self.last1 = data
        self.line += 1
        if self.partition:
            data = data.partition(self.partition)[0]
        if isinstance(self.__cb_next, Callable):
            self.__cb_next(self, data)
        return data

    @property
    def peek_next(self) -> str:
        data = self.next
        self.revert(1)
        return data

    def revert(self, count: int = 1) -> None:
        if count > len(self.last):
            raise ValueError(
                f"Revert count {count} is greater than last line count {len(self.last)}"
            )
        for _ in range(count):
            self.__reg.put(self.last1)
            self.last.pop()
            self.last1 = self.last[-1]
            self.line -= 1

    def __next__(self) -> str:
        return self.next

    def __iter__(self) -> "LineIterator":
        return self



from typing import List, Optional, Union, Sequence
from subprocess import Popen as _Popen, PIPE, TimeoutExpired, CompletedProcess
from abc import ABC, abstractmethod
from queue import Queue
from threading import Thread, Event
from time import sleep
import sys

from .string import startswith

if sys.platform == "win32":
    from select import select
else:
    from fcntl import F_GETFL, F_SETFL, fcntl
    from os import O_NONBLOCK

__all__ = [
    "Popen",
    "Disposable",
]


class Popen(_Popen):
    """
    用于交互式进程。
    """

    def __init__(self, args: Union[Sequence, str], verbose: bool = False) -> None:
        self._config = {}
        self.verbose = verbose
        if isinstance(args, str):
            shell = True
        elif isinstance(args, Sequence):
            shell = False
            args = tuple(map(str, args))
        else:
            raise TypeError("args must be a sequence or a string")
        super().__init__(
            args,
            shell=shell,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            text=True,
            bufsize=1,
        )
        if self.poll() is not None:
            raise ChildProcessError("Process failed to start")

        def reader(self: __class__, handleType: str) -> None:
            def oneline(res: str):
                if res != "":
                    queue.put(res)
                    if self.verbose:
                        end = "" if res.endswith("\n") else "\n"
                        print(f"{handleType}: {res}", end=end)

            handle = getattr(self, handleType)
            queue = getattr(self, f"{handleType}_queue")

            if sys.platform == "linux":
                flags = fcntl(handle, F_GETFL)
                fcntl(handle, F_SETFL, flags | O_NONBLOCK)

            while True:
                if self.thread_event.is_set():
                    break
                if handle.closed:
                    break

                if sys.platform == "win32":
                    read, _, _ = select([handle], [], [], 0.1)
                    if not read:
                        continue

                if self.poll() is not None:
                    map(oneline, handle.readlines())
                    break
                oneline(handle.readline())
            return

        self._config["stdout_queue"] = Queue()
        self._config["stderr_queue"] = Queue()
        self._config["thread_event"] = Event()
        self._config["stdout_thread"] = Thread(
            target=reader, args=(self, "stdout"), daemon=True
        )
        self._config["stderr_thread"] = Thread(
            target=reader, args=(self, "stderr"), daemon=True
        )
        self.stdout_thread.start()
        self.stderr_thread.start()

    @property
    def verbose(self) -> bool:
        return self._config["verbose"]

    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._config["verbose"] = True if value else False

    @property
    def stdout_queue(self) -> Queue:
        return self._config["stdout_queue"]

    @property
    def stderr_queue(self) -> Queue:
        return self._config["stderr_queue"]

    @property
    def thread_event(self) -> Event:
        return self._config["thread_event"]

    @property
    def stdout_thread(self) -> Thread:
        return self._config["stdout_thread"]

    @property
    def stderr_thread(self) -> Thread:
        return self._config["stderr_thread"]

    def stop(self, timeout: int = 300) -> None:
        if self.poll() is not None:
            return
        self.thread_event.set()
        self.stdout_thread.join()
        self.stderr_thread.join()
        if self.stdin and not self.stdin.closed:
            self.stdin.close()
        if self.stdout and not self.stdout.closed:
            self.stdout.close()
        if self.stderr and not self.stderr.closed:
            self.stderr.close()
        if self.poll() is None:
            self.terminate()
            self.wait(timeout=timeout)
            if self.poll() is None:
                self.kill()

    def wait(self, *args, **kwargs):
        returncode = None
        try:
            returncode = super().wait(*args, **kwargs)
        except TimeoutExpired:
            pass
        return returncode

    def send(self, *args: Union[str, int, float]) -> str:
        cmd = " ".join(map(str, args))
        stdin = self.stdin
        if stdin is None or stdin.closed:
            raise BrokenPipeError("stdin is closed")
        stdin.write((cmd + "\n"))
        stdin.flush()
        return cmd

    def recv(
        self,
        targets: Union[str, Sequence[str]],
        heads: Optional[list] = None,
        targets_regex: bool = False,
        targets_case: bool = False,
        heads_regex: bool = False,
        heads_case: bool = False,
    ) -> List[str]:
        if targets is None:
            raise ValueError("targets is None")
        elif isinstance(targets, Sequence):
            targets = list(targets)
        else:
            targets = [targets]
        if len(targets) < 1:
            raise ValueError("targets is empty")

        if heads is not None:
            if isinstance(heads, str):
                heads = [heads]
            elif isinstance(heads, list):
                pass
            else:
                raise ValueError("heads is not str or list")

        q = self.stdout_queue

        if heads is not None:
            while True:
                if q.empty():
                    sleep(0.1)
                    continue
                line = q.queue[0]
                if startswith(line, heads, regex=heads_regex, case=heads_case):
                    break
                q.get(timeout=1)
                q.task_done()

        result = []
        while True:
            if q.empty():
                sleep(0.1)
                continue
            line = q.get()
            result.append(line)
            q.task_done()
            if startswith(line, targets, regex=targets_regex, case=targets_case):
                break
        return result


class Disposable(ABC):
    """
    一次性进程，用于快捷封装自定义进程。
    """

    __slots__ = ("__config",)

    def __init__(self) -> None:
        self.__config = {
            "stdout": None,
            "stderr": None,
            "started": False,
            "process": None,
            "returncode": None,
        }

    @property
    def stdout(self) -> Optional[str]:
        return self.__config["stdout"]

    @property
    def stderr(self) -> Optional[str]:
        return self.__config["stderr"]

    @property
    def started(self) -> bool:
        return self.__config["started"]

    @property
    def process(self) -> _Popen:
        return self.__config["process"]

    @property
    def returncode(self) -> Optional[int]:
        return self.__config["returncode"]

    def start_init(self) -> None:
        """callback order:
        >>> self.start_init()  # here
        >>> if not self.started:
        >>>     self.process_args()
        >>>     if args is not None:
        >>>         self.process_kwargs()
        >>>         self.begin_before()
        >>>         `begin process`
        >>>         self.begin_after()
        """
        pass

    def begin_before(self) -> None:
        """callback order:
        >>> self.start_init()
        >>> if not self.started:
        >>>     self.process_args()
        >>>     if args is not None:
        >>>         self.process_kwargs()
        >>>         self.begin_before()  # here
        >>>         `begin process`
        >>>         self.begin_after()
        """
        pass

    @abstractmethod
    def process_args(self) -> Union[Sequence, str, None]:
        """callback order:
        >>> self.start_init()
        >>> if not self.started:
        >>>     self.process_args()  # here
        >>>     if args is not None:
        >>>         self.process_kwargs()
        >>>         self.begin_before()
        >>>         `begin process`
        >>>         self.begin_after()
        """
        pass

    def process_kwargs(self) -> dict:
        """callback order:
        >>> self.start_init()
        >>> if not self.started:
        >>>     self.process_args()
        >>>     if args is not None:
        >>>         self.process_kwargs()  # here
        >>>         self.begin_before()
        >>>         `begin process`
        >>>         self.begin_after()
        kwargs for Popen, without args, stdout, stderr and shell.
        """
        return {}

    def begin_after(self) -> None:
        """callback order:
        >>> self.start_init()
        >>> if not self.started:
        >>>     self.process_args()
        >>>     if args is not None:
        >>>         self.process_kwargs()
        >>>         self.begin_before()
        >>>         `begin process`
        >>>         self.begin_after()  # here
        """
        pass

    def start(self) -> None:
        self.start_init()
        if not self.started:
            args = self.process_args()
            if args is None:
                raise ValueError("args is None")
            elif isinstance(args, str):
                shell = True
            else:
                shell = False
            kwargs = self.process_kwargs()
            kwargs.update(
                {
                    "args": args,
                    "stdout": PIPE,
                    "stderr": PIPE,
                    "shell": shell,
                }
            )
            self.begin_before()
            self.__config["process"] = _Popen(**kwargs)
            self.__config["started"] = True
            self.begin_after()
        else:
            raise RuntimeError("Process already started")

    def join_before(self) -> None:
        """callback order:
        >>> self.join_before()  # here
        >>> `wait process`
        >>> self.join_after()
        """
        pass

    def join_after(self) -> None:
        """callback order:
        >>> self.join_before()
        >>> `wait process`
        >>> self.join_after()  # here
        """
        pass

    def join(self) -> None:
        proc = self.process
        if self.started and proc is not None and self.returncode is None:
            with proc:
                try:
                    self.join_before()
                    stdout, stderr = proc.communicate()
                    self.join_after()
                except Exception:
                    proc.kill()
                    raise ChildProcessError("some error happened.")
                finally:
                    self.__config["returncode"] = proc.poll()
                    self.__config["stdout"] = stdout.decode()
                    self.__config["stderr"] = stderr.decode()

    def run(self) -> CompletedProcess:
        self.start()
        self.join()
        proc = self.process
        if proc is not None and self.returncode is not None:
            returncode = self.returncode
        else:
            returncode = 1
        return CompletedProcess(proc.args, returncode, self.stdout, self.stderr)



from typing import Optional
import logging

try:
    from rich.logging import RichHandler

    _format = "%(message)s"
    _datefmt = None
    _handlers = [
        RichHandler(
            show_level=True,
            show_path=False,
            log_time_format="%H:%M:%S",
            omit_repeated_times=False,
        )
    ]
except ImportError:
    _format = "%(asctime)s  %(levelname)s    %(message)s"
    _datefmt = "%H:%M:%S"
    _handlers = None

__all__ = [
    "getLogger",
]


def getLogger(name: Optional[str] = None) -> logging.Logger:
    logging.basicConfig(
        format=_format,
        datefmt=_datefmt,
        level=logging.INFO,
        handlers=_handlers,
    )
    return logging.getLogger(name)



from io import TextIOWrapper
from typing import Callable, Iterator, Optional, Union, Any
from dataclasses import dataclass

from ply.lex import (
    LexToken as _LexToken,
    TOKEN,
    Lexer,
    lex,
)


__all__ = [
    "TOKEN",
    "LexToken",
    "MetaLexer",
    "BaseLexer",
    # "tokensToDict",
]


@dataclass
class LexToken(_LexToken):
    lexer: Lexer
    type: str
    value: Any
    lineno: int
    lexpos: int

    def __repr__(self) -> str:
        return super().__repr__()


class MetaLexer:
    def __new__(cls, *args, **kwargs):
        if cls is MetaLexer:
            raise TypeError("BaseLexer cannot be instantiated")
        return super().__new__(cls)

    t_ignore = " \t"

    def t_error(self, t):
        t.lexer.skip(1)

    def __init__(
        self,
        data: Optional[str] = None,
        *,
        cb_input: Optional[Callable] = None,
        cb_token: Optional[Callable] = None,
    ):
        self.lexer = lex(module=self, debug=False)
        self.lexer.callback = {
            "input": cb_input if callable(cb_input) else None,
            "token": cb_token if callable(cb_token) else None,
        }
        if data is not None:
            self.input(data)

    def input(self, data: Union[str, TextIOWrapper]):
        if isinstance(data, str):
            pass
        elif isinstance(data, TextIOWrapper):
            data = data.read()
        else:
            raise TypeError("data must be str or TextIOWrapper")
        self.lexer.input(data)
        if self.lexer.callback["input"] is not None:
            self.lexer.callback["input"](lexer=self.lexer)

    def token(self) -> Optional[LexToken]:
        t = self.lexer.token()
        if t is None:
            return None
        t = LexToken(
            lexer=self.lexer,
            type=t.type,
            value=t.value,
            lineno=t.lineno,
            lexpos=t.lexpos,
        )
        if self.lexer.callback["token"] is not None:
            self.lexer.callback["token"](lexer=self.lexer, token=t)
        return t

    def __iter__(self) -> Iterator[LexToken]:
        return self

    def __next__(self) -> LexToken:
        t = self.token()
        if t is None:
            raise StopIteration
        return t


class BaseLexer(MetaLexer):
    tokens = [
        "ID",  # abc
        "FLOAT",  # 1.23
        "INT",  # 10
    ]

    literals = """()[]{}<>+-*/=~!@#$%^&\\|;:'",.?_"""

    def t_ID(self, t: LexToken):
        r"[a-zA-Z_]\w*"
        return t

    def t_FLOAT(self, t: LexToken):
        r"\d+\.\d+"
        t.value = float(t.value)
        return t

    def t_INT(self, t: LexToken):
        r"\d+(?!\.)"
        t.value = int(t.value)
        return t

    def t_newline(self, t: LexToken):
        r"\n+"
        t.lexer.lineno += len(t.value)



from typing import Any, Dict, Literal, Optional

__all__ = [
    "start",
]


def start(
    export: Optional[Dict[str, Any]] = None,
    shell: Optional[Literal["ipython", "code"]] = None,
) -> None:
    """
    Start the interactive shell.

    Parameters
    ---
        export : A dict of variables to be exported to the interactive shell.
    """
    try:
        if shell == "code":
            raise ImportError
        from IPython import start_ipython

        start_ipython(argv=[], user_ns=export, display_banner=False)
    except ImportError:
        if shell == "ipython":
            raise ImportError("IPython is not installed.")
        from code import interact

        interact(local=export, banner="")



from pathlib import Path
from setuptools import setup, find_packages
import sys

sys.path.insert(0, ".")

import icutk

VERSION = icutk.__version__
DESCRIPTION = "Integrated Circuit Utilities (kit)"
PROJECT = "icutk"
AUTHOR = "YEUNGCHIE"

setup(
    name=PROJECT,
    version=VERSION,
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=Path("README.md").read_text("utf-8"),
    long_description_content_type="text/markdown",
    url=f"https://github.com/{AUTHOR.lower()}/{PROJECT}",
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.7",
    install_requires=Path("requirements.txt").read_text("utf-8").split(),
    keywords=[
        PROJECT,
        "ic",
        "utils",
        "kit",
        "linux",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
    ],
)



from typing import Dict

__all__ = [
    "getSubClassDict",
]


def getSubClassDict(cls, cls_dict: Dict[str, object] = {}) -> Dict[str, object]:
    for subclass in cls.__subclasses__():
        cls_dict[subclass.__name__] = subclass
        getSubClassDict(subclass, cls_dict)
    return cls_dict



from typing import Union
from pathlib import Path
import os

__all__ = [
    "abspath",
    "realpath",
    "expandpath",
]


def abspath(path: Union[str, Path]) -> Path:
    return Path(path).absolute()


def realpath(path: Union[str, Path]) -> Path:
    return Path(path).resolve()


def expandpath(path: Union[str, Path]) -> Path:
    return Path(os.path.expanduser(os.path.expandvars(str(path))))
